/*********************************************************************
 * Software License Agreement (BSD License)
 *
 *  Copyright (c) 2022, Kassow Robots
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   * Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *   * Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *   * Neither the name of the Kassow Robots nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 *  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 *  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 *  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 *  POSSIBILITY OF SUCH DAMAGE.
 *********************************************************************/

#include <kr2_program_api/api_v1/bundles/custom_device.h>
#include <algorithm>


// Concatenate arrays at compilation time
template <typename T, std::size_t N1, std::size_t N2, std::size_t N3>
constexpr std::array<T, N1 + N2 + N3> concat(std::array<T, N1> one, std::array<T, N2> two, std::array<T, N3> three)
{
    std::array<T, N1 + N2+ N3> result{};
    std::size_t index = 0;

    for (auto& el : one) {
        result[index] = std::move(el);
        ++index;
    }
    for (auto& el : two) {
        result[index] = std::move(el);
        ++index;
    }
    for (auto& el : three) {
        result[index] = std::move(el);
        ++index;
    }

    return result;
}

std::string runCommand(std::string cmd)
{
    char buffer[128];
    std::string result = "";
    FILE* pipe = popen(cmd.c_str(), "r");
    
    if (!pipe) throw std::runtime_error("popen() failed!");
    
    try {
        while (!feof(pipe)) {
            if (fgets(buffer, 128, pipe) != NULL)
            result += buffer;
        }
    } catch (...) {
        
    }
    
    pclose(pipe);

    return result;
}


const std::array<std::uint32_t, 4> output_relay_duid_ = {0x00008101, 0x00008102, 0x00008103, 0x00008104};                                                 // [RELAY     x4]
const std::array<std::uint32_t, 8> output_iob_duid_   = {0x00008111, 0x00008112, 0x00008113, 0x00008114, 0x00008115, 0x00008116, 0x00008117, 0x00008118}; // [DIGITAL   x8]
const std::array<std::uint32_t, 4> output_tb_duid_    = {0x00308101, 0x00308102, 0x00308103, 0x00308104};                                                 // [ToolBoard x4]
const std::array<std::uint32_t, 16> output_rdt_duid_ = concat(output_relay_duid_, output_iob_duid_, output_tb_duid_);

enum { KR2_TB_CONF_OUT_DIG_VOLTAGE_12V  = 0x0801000C };
enum { KR2_TB_CONF_OUT_DIG_VOLTAGE_24V  = 0x08010018 };


    //
    // Frame ID
    //
    /**
    * @brief Selector from which System Frame retrieve values.
    * 
    */

enum EFrameID {
    //! Robot Base Frame
    ROBOT_BASE_FRAME = 0, 
    //! Robot Link 1 Frame
    ROBOT_LINK1_FRAME = 1,
    //! Robot Link 2 Frame
    ROBOT_LINK2_FRAME = 2,
    //! Robot Link 3 Frame
    ROBOT_LINK3_FRAME = 3,  
    //! Robot Link 4 Frame
    ROBOT_LINK4_FRAME = 4,  
    //! Robot Link 5 Frame
    ROBOT_LINK5_FRAME = 5,  
    //! Robot Link 6 Frame
    ROBOT_LINK6_FRAME = 6,  
    //! TFC
    TFC_FRAME = 7,
    //! TCP
    TCP_FRAME = 8,
    //! Last Target Pose
    LAST_TARGET_POSE_FRAME = 9,
    //! Adaptive Reference Target Offset
    ADAPTIVE_REF_TARGET_OFFSET_FRAME = 10,
    //! Trajectory Reference Point
    TRAJECTORY_REF_POINT_FRAME = 11
};
std::map<unsigned int, unsigned int> frame_id_map_ =
    {{ROBOT_BASE_FRAME, kr2rc_api::Model::SysId::FRAME_ROBOT_BASE}, /*{ROBOT_LINK1_FRAME, kr2rc_api::Model::SysId::FRAME_ROBOT_LINKS[0]},
     {ROBOT_LINK2_FRAME, kr2rc_api::Model::SysId::FRAME_ROBOT_LINKS[1]}, {ROBOT_LINK3_FRAME, kr2rc_api::Model::SysId::FRAME_ROBOT_LINKS[2]},
     {ROBOT_LINK4_FRAME, kr2rc_api::Model::SysId::FRAME_ROBOT_LINKS[3]}, {ROBOT_LINK5_FRAME, kr2rc_api::Model::SysId::FRAME_ROBOT_LINKS[4]},
     {ROBOT_LINK6_FRAME, kr2rc_api::Model::SysId::FRAME_ROBOT_LINKS[5]}, */ {TFC_FRAME, kr2rc_api::Model::SysId::FRAME_ROBOT_FLANGE},
     {TCP_FRAME, kr2rc_api::Model::SysId::FRAME_ROBOTX_TCP}, {LAST_TARGET_POSE_FRAME, kr2rc_api::Model::SysId::FRAME_TRAJECTORY_TARGET}/*,
     {ADAPTIVE_REF_TARGET_OFFSET_FRAME, kr2rc_api::Model::SysId::FRAME_ROBOT_EOA_ARTO}, {TRAJECTORY_REF_POINT_FRAME, kr2rc_api::Model::SysId::FRAME_ROBOT_EOA_REF_TRAJECTORY}*/
     };
    //
    // Frame values
    //
    /**
    * @brief Selector which Frame values to set.
    * 
    */
enum EFrameValue {

    // Values Types

    //! Rotation and Position coordinates ([x, y, z, r, p, y])
    // World Frame Reference
    POSE_VAL_REF_WF = 0,
    // TFC Reference
    POSE_VAL_REF_TFC = 1,
};
std::map<unsigned int, unsigned int> frame_val_map_ =
    {{POSE_VAL_REF_WF, kr2rc_api::Model::SysId::FRAME_WORLD}, {POSE_VAL_REF_TFC, kr2rc_api::Model::SysId::FRAME_ROBOT_FLANGE}
    };
    //
    // Load ID
    //
    /**
     * @brief System Load to set values.
     * 
     */
enum ELoadID {
    //! LOAD1
    LOAD1 = 0,
    //! LOAD2
    LOAD2 = 1,
};
std::map<unsigned int, unsigned int> load_id_map_ =
    {{LOAD1, kr2rc_api::Load::SysId::LOAD_TOOL}, {LOAD2, kr2rc_api::Load::SysId::LOAD_PAYLOAD}
    };
    //
    // CPU State
    //
    /**
     * @brief CPUState to set values.
     * 
     */
        enum ECPUStateValue {
            //! Package ID 0 temperature (average)
            PACKAGE_ID0_TEMP = 0,  
            //! Core 0 temperature
            CORE_0_TEMP  = 1,  
            //! Core 1 temperature
            CORE_1_TEMP  = 2,
            //! Core 2 temperature
            CORE_2_TEMP  = 3,
            //! Core 3 temperature
            CORE_3_TEMP  = 4,
        };
std::vector<unsigned int> cpu_state_id_map_ =
    {PACKAGE_ID0_TEMP, CORE_0_TEMP, CORE_1_TEMP, CORE_2_TEMP, CORE_3_TEMP
    };

enum EAlarmState {
    //! Halt
    HALT = 0,
    //! Suspended
    SUSPENDED = 1,
    //! Init
    INIT = 2,
    //! Paused
    PAUSED = 3,
    //! CBunEvent (Stats)
    CBUN_EVENT = 4
};


namespace kswx_kr2_kord {

class UDPKord : public kr2_bundle_api::CustomDevice {

    public:
        UDPKord(boost::shared_ptr<kr2_program_api::ProgramInterface> api,  const boost::property_tree::ptree &xml_bundle_node);
        virtual ~UDPKord();

        CBUN_PCALL sshExportLogs(const std::string &host, const std::string &username, const std::string &path);
        CBUN_PCALL sshExportLogsIni(const std::string &fname);

        // BundleClass interface
    protected:
        virtual int onCreate() override;
        virtual int onDestroy() override;
        virtual int onBind() override;
        virtual int onUnbind() override;

        // CustomDevice interface
    protected:
        virtual CBUN_PCALL onActivate(const boost::property_tree::ptree &param_tree) override;
        virtual CBUN_PCALL onDeactivate() override;
        virtual CBUN_PCALL onMount(const boost::property_tree::ptree &param_tree) override;
        virtual CBUN_PCALL onUnmount() override;

    private:
        class Internals;
        boost::shared_ptr<Internals> his_;
};

} // namespace kswx_kr2_kord
